/*
 * nash-network.c
 *
 * Simple network bring-up code for nash.
 * It currently uses libpump and looks a lot like the network code in
 * anaconda
 *
 * Jeremy Katz <katzj@redhat.com>
 *
 * Copyright 2006 Red Hat, Inc.
 *
 * This software may be freely redistributed under the terms of the GNU
 * General Public License, version 2.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <errno.h>
#include <resolv.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <popt.h>
#include <pump.h>

#include <net/if.h>

#include <asm/types.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <sys/socket.h>

#include <errno.h>

#include <sys/ioctl.h>
#include <linux/sockios.h>

#include <nash.h>
#include "util.h"

/* returns 1 for link, 0 for no link, -1 for unknown */
int get_link_status(char *ifname);

#define NUM_LINK_CHECKS 5
static int waitForLink(char * dev) {
    int tries = 0;

    /* try to wait for a valid link -- if the status is unknown or
     * up continue, else sleep for 1 second and try again for up
     * to five times */
    nashLogger(_nash_context, NASH_NOTICE, "waiting for link... ");
    while (tries < NUM_LINK_CHECKS) {
      if (get_link_status(dev) != 0)
            break;
        sleep(1);
        tries++;
    }
    nashLogger(_nash_context, NASH_NOTICE, "%d seconds.\n", tries);
    if (tries < NUM_LINK_CHECKS)
        return 0;
    nashLogger(_nash_context, NASH_WARNING, "no network link detected on %s\n", dev);
    return 1;
}

static void writeResolvConf(struct pumpNetIntf intf) {
    char * filename = "/etc/resolv.conf";
    FILE * f;
    char buf[INET6_ADDRSTRLEN+1];
    ip_addr_t *addr;
    int i;

    if (!(intf.set & PUMP_NETINFO_HAS_DOMAIN) && !intf.numDns)
        return;

    f = fopen(filename, "w");
    if (!f) {
        eprintf("Cannot create %s: %s\n", filename, strerror(errno));
        return;
    }

    if (intf.set & PUMP_NETINFO_HAS_DOMAIN)
        fprintf(f, "search %s\n", intf.domain);

    for (i = 0; i < intf.numDns; i++) {
        addr = &(intf.dnsServers[i]);
        inet_ntop(addr->sa_family, IP_ADDR(addr), buf, INET6_ADDRSTRLEN);
        fprintf(f, "nameserver %s\n", buf);
    }

    fclose(f);

    res_init();         /* reinit the resolver so DNS changes take affect */

    return;
}

static void nashNetLogger(void * arg, int priority, char * fmt, va_list va) {
    nash_log_level loglevel = NASH_NOTICE;

    if (priority <= LOG_ERR)
        loglevel = NASH_ERROR;
    else if (priority <= LOG_WARNING)
        loglevel = NASH_WARNING;

    nashLoggerV(_nash_context, loglevel, fmt, va);
    nashLogger(_nash_context, loglevel, "\n");
}

static inline int nashPton(char * addr_str, ip_addr_t *ret) {
    struct in_addr addr;
    struct in6_addr addr6;

    if (!addr_str)
        return 0;

    if (inet_pton(AF_INET, addr_str, &addr)) {
	*ret = ip_addr_in(&addr);
	return 1;
    } else if (inet_pton(AF_INET6, addr_str, &addr6)) {
	*ret = ip_addr_in6(&addr6);
	return 1;
    }

    return 0;
}

int nashNetworkCommand(int argc, char *argv[]) {
    char * bootProto = NULL, * dev = NULL, * dhcpclass = NULL, * ethtool = NULL, * hostname = NULL;
    char * gateway = NULL, * ip = NULL, * nameserver = NULL, * netmask = NULL, * dns = NULL, * domain = NULL;
    int mtu = 0, rc;
    char * err = NULL;
    struct pumpNetIntf intf;
    struct in_addr addr;
    poptContext optCon;
    struct poptOption netOptions[] = {
        { "bootproto", '\0', POPT_ARG_STRING, &bootProto, 0, NULL, NULL },
        { "device", '\0', POPT_ARG_STRING, &dev, 0, NULL, NULL },
        { "dhcpclass", '\0', POPT_ARG_STRING, &dhcpclass, 0, NULL, NULL },
        { "dns", '\0', POPT_ARG_STRING, &dns, 0, NULL, NULL },
        { "domain", '\0', POPT_ARG_STRING, &domain, 0, NULL, NULL },
        { "gateway", '\0', POPT_ARG_STRING, &gateway, 'g', NULL, NULL },
        { "ip", '\0', POPT_ARG_STRING, &ip, 'i', NULL, NULL },
        { "nameserver", '\0', POPT_ARG_STRING, &nameserver, 'n', NULL, NULL },
        { "netmask", '\0', POPT_ARG_STRING, &netmask, 'm', NULL, NULL },
        { "ethtool", '\0', POPT_ARG_STRING, &ethtool, 0, NULL, NULL },
        { "mtu", '\0', POPT_ARG_INT, &mtu, 0, NULL, NULL },
        { "hostname", '\0', POPT_ARG_STRING, &hostname, 0, NULL, NULL },
        { NULL, 0, 0, NULL, 0, NULL, NULL }
    };
    
    optCon = poptGetContext(NULL, argc, (const char **) argv, 
                            netOptions, 0);    

    while ((rc = poptGetNextOpt(optCon)) > 0) {}
    if (rc < -1) {
        nashLogger(_nash_context, NASH_ERROR, "ERROR: Bad argument to network command\n");
        return 1;
    }

    if (ethtool != NULL) {
        /* FIXME */
        nashLogger(_nash_context, NASH_WARNING, "WARNING: ethtool options not currently handled\n");
    }

    memset(&intf,'\0', sizeof(intf));

    if (mtu) {
        intf.mtu = mtu;
        intf.set |= PUMP_INTFINFO_HAS_MTU;
    }

    if (hostname != NULL) {
        intf.hostname = hostname;
        intf.set |= PUMP_NETINFO_HAS_HOSTNAME;
    }
    
    if (domain != NULL) {
        intf.domain = domain;
        intf.set |= PUMP_NETINFO_HAS_DOMAIN;
    }

    if (dns) {
        char *c, *buf = strdup(dns);
        
        c = strtok(buf, ",");
        while ((intf.numDns < MAX_DNS_SERVERS) && (c != NULL)) {
            if (nashPton(c, &intf.dnsServers[intf.numDns]))
                intf.numDns++;
            c = strtok(NULL, ",");
        }
        if (intf.numDns)
            intf.set |= PUMP_NETINFO_HAS_DNS;
    }

    if (dev == NULL)
        dev = strdup("eth0");

    strncpy(intf.device, dev, 9);

    if ((bootProto != NULL) && (!strncmp(bootProto, "dhcp", 4))) {
        waitForLink(dev);
        nashLogger(_nash_context, NASH_NOTICE, "Sending request for IP information through %s\n", dev);
        pumpDhcpClassRun(&intf, NULL,
                         dhcpclass ? dhcpclass : "nash",
                         DHCPv6_DISABLE, DHCP_USE_LEASE_DATABASE,
                         45, nashNetLogger, LOG_INFO);
    } else { /* static IP.  hope enough is specified! */
        if (nashPton(ip, &intf.ip))
            intf.set |= PUMP_INTFINFO_HAS_IP;
        if (nashPton(netmask, &intf.netmask))
            intf.set |= PUMP_INTFINFO_HAS_NETMASK;
        if (nashPton(gateway, &intf.gateway))
            intf.set |= PUMP_NETINFO_HAS_GATEWAY;

        /* FIXME: what about IPv6 ? */
        if (intf.set & PUMP_INTFINFO_HAS_NETMASK &&
            intf.netmask.sa_family == AF_INET) {
            addr.s_addr = IP_ADDR_IN(&intf.ip)->s_addr & IP_ADDR_IN(&intf.netmask)->s_addr;
            intf.network = ip_addr_in(&addr);
            addr.s_addr = IP_ADDR_IN(&intf.network)->s_addr | ~IP_ADDR_IN(&intf.netmask)->s_addr;
            intf.broadcast = ip_addr_in(&addr);
            intf.set |= PUMP_INTFINFO_HAS_NETWORK | PUMP_INTFINFO_HAS_BROADCAST;
        }
    }

    err =  pumpSetupInterface(&intf);
    if (err) {
        nashLogger(_nash_context, NASH_ERROR, "ERROR: Interface setup failed: %s\n", err);
        return 1;
    }
    writeResolvConf(intf);

    sleep(2);
    return 0;
}

struct nl {
    int fd;
    struct sockaddr_nl addr;
    int seq;
};

static int
rtnl_wilddump_request(struct nl *nl, int family, int type)
{
    struct sockaddr_nl nladdr;
    struct {
        struct nlmsghdr nh;
        struct rtgenmsg g;
    } req;
    int status;

    memset(&nladdr, '\0', sizeof(nladdr));
    nladdr.nl_family = AF_NETLINK;

    memset(&req, '\0', sizeof(req));
    req.nh.nlmsg_len = sizeof(req);
    req.nh.nlmsg_flags = NLM_F_DUMP | NLM_F_REQUEST ;
    req.nh.nlmsg_type = type;
    req.nh.nlmsg_seq = ++nl->seq;
    req.nh.nlmsg_pid = 0;
    req.g.rtgen_family = family;

    status = sendto(nl->fd, &req, sizeof(req), 0, (struct sockaddr *)&nladdr,
            sizeof(nladdr));
    if (status < 0)
        return -1;

    return 0;
}

struct ifinfo {
    int index;
    char *name;
    char *mac;
    struct ifinfo *next;
};

static int
rtnl_dump_recv(struct nl *nl, struct ifinfo **iftable)
{
    struct sockaddr_nl nladdr;
    struct iovec iov;
    struct msghdr msg = {
        .msg_name = &nladdr,
        .msg_namelen = sizeof(nladdr),
        .msg_iov = &iov,
        .msg_iovlen = 1,
    };
    char buf[16384];
    struct ifinfo *ifentry = NULL;

    iov.iov_base = buf;
    while (1) {
        int status;
        struct nlmsghdr *h;

        memset(buf, '\0', sizeof(buf));
        iov.iov_len = sizeof(buf);
        status = recvmsg(nl->fd, &msg, 0);

        if (status < 0) {
            if (errno == EINTR)
                continue;
            return -1;
        }

        if (status == 0)
            return -1;

        h = (struct nlmsghdr *)buf;
        while (NLMSG_OK(h, status)) {
            struct rtattr *rta;
            struct ifinfomsg *ifi;
            int len;

            if (h->nlmsg_type == NLMSG_DONE)
                return 0;

            if (nladdr.nl_pid != 0 || h->nlmsg_pid != nl->addr.nl_pid ||
                    h->nlmsg_seq != nl->seq)
                goto skip_it;

            if (h->nlmsg_type == NLMSG_ERROR)
                goto skip_it;

            if (h->nlmsg_type == RTM_NEWLINK) {
                ifentry = calloc(1, sizeof(*ifentry));

                ifi = NLMSG_DATA(h);
                ifentry->index = ifi->ifi_index;
                len = IFLA_PAYLOAD(h);
                rta = IFLA_RTA(ifi);
                while (RTA_OK(rta, len)) {
                    switch(rta->rta_type) {
                        case IFLA_IFNAME:
                            ifentry->name = strdup(RTA_DATA(rta));
                            break;
                        case IFLA_ADDRESS: {
                            int alen = RTA_PAYLOAD(rta);
                            int i;
                            char *mac = calloc(alen * 2 + alen, sizeof(char));
                            char *data, *octet;

                            data = RTA_DATA(rta);
                            octet = mac;

                            for (i = 0; i < alen; i++) {
                                sprintf(octet, "%02x:", (unsigned char)data[i]);
                                octet += 3;
                            }
                            mac[i*2+i-1] = '\0';
                            ifentry->mac = mac;
                            break;
                        }
                        default:
                            break;
                    }
                    rta = RTA_NEXT(rta,len);
                }
                ifentry->next = *iftable;
                *iftable = ifentry;
            }
            
skip_it:
            h = NLMSG_NEXT(h, status);
        }
    }

    return 0;
}

static int
get_ifinfo(struct ifinfo **table)
{
    struct nl nl;
    socklen_t socklen;
    int n;
    int rc;

    memset(&nl, '\0', sizeof(nl));
    nl.addr.nl_family = AF_NETLINK;
    nl.seq = time(NULL);

    nl.fd = socket(PF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
    if (nl.fd < 0) {
        fprintf(stderr, "socket() failed: %m\n");
        exit(1);
    }

    n = 16384;
    socklen = sizeof(n);
    setsockopt(nl.fd, SOL_SOCKET, SO_SNDBUF, &n, socklen);
    n = 16384;
    setsockopt(nl.fd, SOL_SOCKET, SO_RCVBUF, &n, socklen);
    bind(nl.fd, (struct sockaddr *)&nl.addr, sizeof(nl.addr));

    memset(&nl.addr, 0, sizeof(nl.addr));
    socklen = sizeof(nl.addr);
    getsockname(nl.fd, (struct sockaddr *)&nl.addr, &socklen);

    rc = rtnl_wilddump_request(&nl, AF_UNSPEC, RTM_GETLINK);
    if (rc >= 0)
        rc = rtnl_dump_recv(&nl, table);

    close(nl.fd);
    return rc;
}

static void
free_ifinfo(struct ifinfo *table)
{
    struct ifinfo *ifentry = table;

    while (ifentry) {
        struct ifinfo *next = ifentry->next;

        if (ifentry->name)
            free(ifentry->name);
        if (ifentry->mac)
            free(ifentry->mac);
        free(ifentry);
        ifentry = next;
    }
}

static int
get_ctl_fd(void)
{
	int s_errno;
	int fd;

	fd = socket(PF_INET, SOCK_DGRAM, 0);
	if (fd >= 0)
		return fd;
	s_errno = errno;
	fd = socket(PF_PACKET, SOCK_DGRAM, 0);
	if (fd >= 0)
		return fd;
	fd = socket(PF_INET6, SOCK_DGRAM, 0);
	if (fd >= 0)
		return fd;
	errno = s_errno;
	perror("Cannot create control socket");
	return -1;
}

static int
do_changename(const char *dev, const char *newdev)
{
	struct ifreq ifr;
	int fd;
	int err, en;

	strncpy(ifr.ifr_name, dev, IFNAMSIZ);
	strncpy(ifr.ifr_newname, newdev, IFNAMSIZ);
	fd = get_ctl_fd();
	if (fd < 0)
		return -1;
	err = ioctl(fd, SIOCSIFNAME, &ifr);
        en = errno;
	close(fd);
        errno = en;
        return err;
}

int
renameNetworkInterface(char *hwaddr, char *ifname)
{
    struct ifinfo *ifinfo = NULL, *ifentry;
    int rc = 0;

    int collision = 0;
    char *oldname = NULL;

    if (get_ifinfo(&ifinfo) >= 0) {
        for (ifentry = ifinfo; ifentry; ifentry = ifentry->next) {
            if (!strcasecmp(hwaddr, ifentry->mac)) {
                if (!strcmp(ifname, ifentry->name))
                    break;
                oldname = ifentry->name;
            } else if (!strcmp(ifname, ifentry->name)) {
                collision = 1;
            }
        }
        if (collision)
            rc = do_changename(ifname, "nashdevname");
        if (oldname) {
            if (rc >= 0)
                rc = do_changename(oldname, ifname);
            if (collision) {
                if (rc >= 0)
                    do_changename("nashdevname", oldname);
                else
                    do_changename("nashdevname", ifname);
            }
        }

        free_ifinfo(ifinfo);
    }

    return rc;
}

